package anynt.environment;


import java.util.Random;
import anynt.environment.settings.RewardsProvider;
import anynt.object.Agent;
import anynt.object.EvaluableAgent;
import anynt.object.Interaction;
import anynt.object.RewardsGeneratorAgent;
import anynt.space.Cell;
import anynt.space.Observation;
import anynt.space.Observation_L;
import anynt.space.Space;


/**
 * Extends an environment to represent a Lambda Environment.
 * The Lambda Environment contains one space where agents will interact.
 * Two agents are added "Good" and "Evil", which will provide the rewards.
 * 
 * @author Javier
 * @since 11/04/2011
 */
public class LambdaEnvironment extends Environment
{
	/**
	 * Space of the environment
	 */
	protected Space space;
	/**
	 * Good agent, it gives good rewards
	 */
	protected RewardsGeneratorAgent goodAgent;
	/**
	 * Evil agent, it gives bad rewards
	 */
	protected RewardsGeneratorAgent evilAgent;
	/**
	 * Number of interactions to realocate good and evil
	 */
	private int interactionsToRelocateRewardsGenerators;
	/**
	 * Factor that rewards are divided on each iteration
	 */
	private int decreaseFactor;
	/**
	 * Moment to consume rewards
	 */
	private boolean consumeRewardsBeforeRewardersUpdateCellRewards;
	/**
	 * Tells if rewards are consumed/deleted from the space
	 */
	private boolean evaluableAgentConsumeRewards;
	/**
	 * It tells how rewards are provided/shared
	 */
	private RewardsProvider rewardsProvider;


	// Constructors
	public LambdaEnvironment()
	{
		goodAgent = new RewardsGeneratorAgent("Good", +1.0);
		addAgent(goodAgent);
		evilAgent = new RewardsGeneratorAgent("Evil", -1.0);
		addAgent(evilAgent);

		decreaseFactor = 2;
		consumeRewardsBeforeRewardersUpdateCellRewards = true;
		evaluableAgentConsumeRewards = true;
	}

	// Environment overrides
	public void checkConsistency()
	{
		super.checkConsistency();
		if (space == null)
			throw new RuntimeException("LambdaEnvironment space is not defined");
		space.checkConsistency();
		if (goodAgent == null)
			throw new RuntimeException("LambdaEnvironment Good Agent is not defined");
		goodAgent.checkConsistency();
		if (evilAgent == null)
			throw new RuntimeException("LambdaEnvironment Evil Agent is not defined");
		evilAgent.checkConsistency();
		if (agents.size() == 2)
			throw new RuntimeException("LambdaEnvironment only has good and evil agents");
		if (rewardsProvider == null)
			throw new RuntimeException("LambdaEnvironment Rewards Provider is not defined");
		if (interactionsToRelocateRewardsGenerators < 0)
			throw new RuntimeException("Interactions to relocate reward generators must be positive or 0");
		if (decreaseFactor <= 0)
			throw new RuntimeException("Decrease factor must be positive");
	}
	/**
	 * Reset data structures
	 */
	public void reset()
	{
		super.reset();

		space.reset();
	}

	/**
	 * Prepares the Environment to be launched
	 */
	protected final void prepareEnvironment()
	{
		initialLocateAgents();

		// Set a reward of 0.0 in all cells
		for (int index = 1; index <= space.getNumberOfCells(); index++)
		{
			Cell cell = space.getCell(index);
			cell.setReward(0.0);
		}

		dropRewards();
	}
	/**
	 * Locates the agents on the environment for the first interaction
	 */
	protected void initialLocateAgents()
	{
		int numCells = space.getNumberOfCells();

		// Locate randomly all the agents
		for (Agent agent : agents)
			locateAgent(agent, new Random().nextInt(numCells) + 1);

		// Relocates the Good and Evil Agents if they where in the same cell
		while (goodAgent.getLocation().equals(evilAgent.getLocation()))
		{
			locateAgent(goodAgent, new Random().nextInt(numCells) + 1);
			locateAgent(evilAgent, new Random().nextInt(numCells) + 1);
		}
	}
	/**
	 * Select the Observation to show to the Agent
	 */
	protected final Observation takeObservation()
	{
		return new Observation_L(space.clone());
	}
	/**
	 * Prepare a new interactions if it's needed
	 */
	protected void prepareInteraction(int interactionNumber)
	{
		// If reward generators may be relocated on this interaction, relocate them
		if (interactionsToRelocateRewardsGenerators != 0 && interactionNumber!= 0
		&& interactionNumber % interactionsToRelocateRewardsGenerators == 0)
			relocateRewardsGeneratorsAgents();
	}
	/**
	 * Relocate the Good and Evil Agents
	 */
	protected void relocateRewardsGeneratorsAgents()
	{
		int numCells = space.getNumberOfCells();

		locateAgent(goodAgent, new Random().nextInt(numCells) + 1);
		locateAgent(evilAgent, new Random().nextInt(numCells) + 1);

		// Relocates the Good and Evil Agents if they where in the same cell
		while (goodAgent.getLocation().equals(evilAgent.getLocation()))
		{
			locateAgent(goodAgent, new Random().nextInt(numCells) + 1);
			locateAgent(evilAgent, new Random().nextInt(numCells) + 1);
		}
	}
	protected void makeAgentsActions(Interaction[] interactions)
	{
		Cell previousGoodCell = goodAgent.getLocation();
		Cell previousEvilCell = evilAgent.getLocation();

		super.makeAgentsActions(interactions);

		ensureRewardsGeneratorsAreNotInTheSameCell(previousGoodCell, previousEvilCell);
	}
	/**
	 * Ensure rewards generator don't fall in the same cell
	 */
	private void ensureRewardsGeneratorsAreNotInTheSameCell(Cell previousGoodCell, Cell previousEvilCell)
	{
		if (goodAgent.getLocation().equals(evilAgent.getLocation()))
			// If the Good Agent didn't move the Evil Agent don't move
			if (goodAgent.getLocation().equals(previousGoodCell))
				locateAgent(evilAgent, previousEvilCell.getNumber());
			// If the Evil Agent didn't move the Good Agent don't move
			else if (evilAgent.getLocation().equals(previousEvilCell))
				locateAgent(goodAgent, previousGoodCell.getNumber());
			// If both Agents had move one of them don't move
			else if (new Random().nextInt(2) == 0)
				locateAgent(goodAgent, previousGoodCell.getNumber());
			else
				locateAgent(evilAgent, previousEvilCell.getNumber());
	}
	/**
	 * Moves the Agent to the new cell by calculating the destination cell depending on the action
	 */
	protected final void makeAction(Agent agent, int[] action)
	{
		Cell newCell = agent.getLocation();
		for (int index = 0; index < action.length; index++)
			newCell = newCell.destinationCell(action[index]);

		space.locateObject(agent, newCell.getNumber());
	}
	/**
	 * Actualize the environment in function of the environment features
	 */
	protected final void actualize()
	{
		// If the rewards are consumed before they are updated
		if (consumeRewardsBeforeRewardersUpdateCellRewards)
		{
			evaluateAgentsActions(interactions);
			consumeRewards();
		}

		// Update the environment with the news rewards
		updateRewardsOnTheSpace();
		dropRewards();

		// If the rewards are consumed after they are updated
		if (!consumeRewardsBeforeRewardersUpdateCellRewards)
		{
			evaluateAgentsActions(interactions);
			consumeRewards();
		}
	}
	/**
	 * Evaluate agents' actions
	 */
	private void evaluateAgentsActions(Interaction[] interactions)
	{
		// Set rewards to the Agents
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);
			// Evaluates the Agent action in the Environment
			double obtainedReward = rewardsProvider.providedReward(agent.getLocation());
			interactions[actualAgent].setReward(obtainedReward);
		}
	}
	/**
	 * Evaluable agents consume rewards if applicable
	 */
	private void consumeRewards()
	{
		if (!evaluableAgentConsumeRewards)
			return;

		for (Agent agent : agents)
			if (agent instanceof EvaluableAgent)
			{
				Cell cell = agent.getLocation();
				cell.setReward(0.0);
			}
	}
	/**
	 * Update the rewards dividing them by the decrease factor
	 */
	private void updateRewardsOnTheSpace()
	{
		// Update the rewards of all cells by dividing by the decrease factor
		for (int index = 1; index <= space.getNumberOfCells(); index++)
		{
			Cell cell = space.getCell(index);
			if (decreaseFactor == Integer.MAX_VALUE)
				cell.setReward(0.0);
			else
				cell.setReward(cell.getReward() / decreaseFactor);
		}
	}

	// Treatment of object attributes
	public Space getSpace()
	{
		return space;
	}
	public void setSpace(Space space)
	{
		this.space = space;

		int numActions = space.getNumberOfActions();

		actions = new int[numActions];
		for (int numAction = 0; numAction < numActions; numAction++)
			actions[numAction] = numAction;
	}
	public RewardsGeneratorAgent getGoodAgent()
	{
		return goodAgent;
	}
	public RewardsGeneratorAgent getEvilAgent()
	{
		return evilAgent;
	}
	public int getInteractionsToRelocateRewarders()
	{
		return interactionsToRelocateRewardsGenerators;
	}
	public void setInteractionsToRelocateRewarders(int interactions)
	{
		interactionsToRelocateRewardsGenerators = interactions;
	}
	public int getDecreaseFactor()
	{
		return decreaseFactor;
	}
	public void setDecreaseFactor(int factor)
	{
		decreaseFactor = factor;
	}
	public boolean getRewardsConsumedBeforeRewardersUpdateCellRewards()
	{
		return consumeRewardsBeforeRewardersUpdateCellRewards;
	}
	public void setRewardsConsumedBeforeRewardersUpdateCellRewards(boolean consumeBefore)
	{
		consumeRewardsBeforeRewardersUpdateCellRewards = consumeBefore;
	}
	public boolean getEvaluableAgentConsumeRewards()
	{
		return evaluableAgentConsumeRewards;
	}
	public void setEvaluableAgentConsumeRewards(boolean consume)
	{
		evaluableAgentConsumeRewards = consume;
	}
	public RewardsProvider getRewardsSharing()
	{
		return rewardsProvider;
	}
	public void setRewardsProvider(RewardsProvider rewardsProvider)
	{
		this.rewardsProvider = rewardsProvider;
	}

	// Object functions
	/**
	 * Locates the agent in the cell with this number
	 */
	protected final void locateAgent(Agent agent, int cellNumber)
	{
		space.locateObject(agent, cellNumber);
	}
	/**
	 * Rewards Generators drop their rewards
	 */
	protected void dropRewards()
	{
		Cell goodCell = goodAgent.getLocation();
		Cell evilCell = evilAgent.getLocation();

		goodCell.setReward(goodAgent.getDropReward());
		evilCell.setReward(evilAgent.getDropReward());
	}
}