package anynt.exercise;


import observer.Subject;

import anynt.environment.Environment;
import anynt.exercise.settings.InteractionType;
import anynt.exercise.settings.StepWaiting;
import anynt.object.EvaluableAgent;
import anynt.object.behaviour.HumanBehaviour;


/**
 * It represent an exercise
 * 
 * @author Javier
 * @since 11/04/2011
 */
public class Exercise extends Subject implements Runnable
{
	// Object attributes
	/**
	 * Environment that will be used in the exercise
	 */
	private Environment environment;
	/**
	 * Agent that will be evaluated in the exercise
	 */
	private EvaluableAgent evaluableAgent;
	/**
	 * Interaction type.
	 * This decides how many interactions/time is the exercise running
	 */
	private InteractionType interactionType;
	/**
	 * Wait between interactions
	 */
	private StepWaiting stepWaiting;

	/**
	 * The exercise is over
	 */
	private boolean finished;
	/**
	 * Thread of the exercise
	 */
	private Thread thread;


	// Runnable overrides
	public void run()
	{
		interact();
	}

	// Treatment of object attributes
	public Environment getEnvironment()
	{
		return environment;
	}
	public void setEnvironment(Environment environment)
	{
		this.environment = environment;
	}
	public EvaluableAgent getEvaluableAgent()
	{
		return evaluableAgent;
	}
	public void setEvaluableAgent(EvaluableAgent evaluableAgent)
	{
		this.evaluableAgent = evaluableAgent;
	}
	public InteractionType getInteractionType()
	{
		return interactionType;
	}
	public void setInteractionType(InteractionType interactionType)
	{
		this.interactionType = interactionType;
	}
	public StepWaiting getStepWaiting()
	{
		return stepWaiting;
	}
	public void setStepWaiting(StepWaiting stepWaiting)
	{
		this.stepWaiting = stepWaiting;
	}
	public boolean hasFinished()
	{
		return finished;
	}

	// Object functions
	/**
	 * Reset the exercise to his initial definition
	 */
	@SuppressWarnings("deprecation")
	public void reset()
	{
		if (thread != null && thread.isAlive())
			thread.stop();
		environment.reset();
	}

	/**
	 * It checks if the exercise is well defined and can be started
	 */
	public void checkConsistency()
	{
		if (environment == null)
			throw new RuntimeException("The Environment of the exercise is not defined");
		environment.checkConsistency();
		if (evaluableAgent == null)
			throw new RuntimeException("The Evaluable Agent of the exercise is not defined");
		evaluableAgent.checkConsistency();
		if (interactionType == null)
			throw new RuntimeException("The Interaction Type of the exercise is not defined");
		if (stepWaiting == null)
			throw new RuntimeException("The Step Waiting of the exercise is not defined");
	}
	/**
	 * Start the exercise
	 * All the parameters must be well instantiated
	 */
	public void start()
	{
		thread = new Thread(this, "Exercise");
		thread.start();
	}
	/**
	 * Waits while the exercise finishes
	 */
	public void join() throws InterruptedException
	{
		thread.join();
	}
	/**
	 * It launchs the interaction between the environment and the agents
	 */
	public void interact()
	{
		finished = false;
		environment.prepare();

		// Loop for the interaction between the Environment and the Agents
		for (interactionType.start(); !interactionType.checkOver(); interactionType.end())
		{
			// Starts the interaction, sending rewards and observations to agents
			environment.beginInteraction(interactionType.getInteractionNumber());

			// Notifies to all the observers that the object has been changed
			notifyObservers();

			// It obtains agents actions
			environment.makeInteraction();

			// If the agent that is been evaluating is not human, then leave a wait time between steps
			if (!(evaluableAgent.getBehaviour() instanceof HumanBehaviour))
				stepWaiting.waitStep();

			// It checks if the interaction is over
			// It serves to check if it is exceeded the time devoted to the exercise
			if (interactionType.checkOver())
				break;

			// It realizes agents' actions and actualizes the environment
			environment.finishInteraction();
		}

		// Send to the agents the last rewards
		environment.finish();
		finished = true;

		
		// Notifies to all the observers that the object has been changed
		notifyObservers();
	}
}