package anynt.space;


import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import anynt.object.Agent;
import anynt.object.Objeto;

import graph.Edge;
import graph.Vertex;


/**
 * It extends a vertex to represent a cell
 * 
 * @author Javier
 * @since 28/03/2011
 */
public class Cell extends Vertex
{
	// Object attributes
	/**
	 * Objects contained in the cell
	 */
	private final ArrayList<Objeto> objects;
	/**
	 * Cell reward
	 */
	private double reward;

	/**
	 * Agents contained in the cell
	 */
	private final ArrayList<Agent> agents;
	/**
	 * Objects contained in the cell that are no agents
	 */
	private final ArrayList<Objeto> noAgents;


	// Constructors
	public Cell(int number)
	{
		super(number);

		objects = new ArrayList<Objeto>();
		noAgents = new ArrayList<Objeto>();
		agents = new ArrayList<Agent>();
		reward = 0.0;
	}

	// Object overrides
	public Cell clone()
	{
		Cell cell = new Cell(getIdentifier());

		for (Objeto object : objects)
			cell.addObject(object);
		cell.reward = reward;

		return cell;
	}

	// Treatment of object attributes
	public int getNumber()
	{
		return getIdentifier();
	}
	public Collection<Objeto> getObjects()
	{
		return objects;
	}
	public void addObject(Objeto object)
	{
		objects.add(object);
		if (object instanceof Agent)
			agents.add((Agent) object);
		else
			noAgents.add(object);
	}
	public void removeObject(Objeto object)
	{
		objects.remove(object);
		agents.remove(object);
		noAgents.remove(object);
	}
	public Collection<Agent> getAgents()
	{
		return agents;
	}
	public Collection<Objeto> getObjectsNoAgents()
	{
		return noAgents;
	}
	public double getReward()
	{
		return reward;
	}
	public void setReward(double reward)
	{
		this.reward = reward;
	}

	// Object function
	/**
	 *  Reset the cell to his initial definition
	 */
	public void reset()
	{
		this.objects.clear();
		this.noAgents.clear();
		this.agents.clear();
		this.reward = 0.0;
	}

	/**
	 *  Returns the available actions that can be done in this Cell
	 */
	public Collection<Integer> getAvailableActions()
	{
		ArrayList<Integer> actions = new ArrayList<Integer>();

		for (Iterator<Edge> iterator = getEdges().iterator(); iterator.hasNext(); )
		{
			Connection connection = (Connection) iterator.next();
			// It takes only the actions that start in this cell
			if (connection.getStartCell() == this)
				actions.add(connection.getAction());
		}

		return actions;
	}
	/**
	 *  Returns the connected cells from this cell
	 */
	public Collection<Cell> getConnectedCells()
	{
		ArrayList<Cell> connectedCells = new ArrayList<Cell>();

		for (Iterator<Vertex> iterator = getAdyacents().iterator(); iterator.hasNext(); )
			connectedCells.add((Cell) iterator.next());

		return connectedCells;
	}
	/**
	 *  Returns the destination Cell once the Action is made
	 *  If the cell does not have this action the returns itself
	 */
	public Cell destinationCell(int action)
	{
		for (Iterator<Edge> iterator = getEdges().iterator(); iterator.hasNext(); )
		{
			Connection connection = (Connection) iterator.next();
			// If this action exists in the cell
			if (connection.getStartCell().equals(this) && connection.getAction() == action)
				return connection.getEndCell();
		}
		return this;
	}
}