package anynt.space;


import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.StringTokenizer;

import anynt.object.Objeto;

import distribution.IDistribution;
import distribution.UniformDistribution;
import distribution.UniversalDistribution;
import exception.CellNotFoundException;
import graph.Edge;
import graph.Graph;
import graph.Vertex;


/**
 * It extends a graph to represent an space
 * 
 * @author Javier
 * @since 28/03/2011
 */
public class Space extends Graph
{
	// Object attributes
	/**
	 * Description of the space in String format
	 */
	private String description;
	/**
	 * Number of actions that has the space
	 */
	private int numberOfActions;
	/**
	 * Minimum number of cells that must be generated
	 */
	private int minCells;
	/**
	 * Maximum number of cells that must be generated
	 */
	private int maxCells;
	/**
	 * Minimum number of actions that must be generated
	 */
	private int minActions;
	/**
	 * Maximum number of actions that must be generated
	 */
	private int maxActions;


	// Constructors
	public Space()
	{
		minCells = 2;
		maxCells = 9;
		minActions = 2;
		maxActions = 9;
	}

	// Object overrides
	public Space clone()
	{
		Space space = new Space();

		// Clones all the vertex
		for (Iterator<Vertex> iterator = vertices.values().iterator(); iterator.hasNext(); )
		{
			Cell cell = (Cell) iterator.next();
			space.addCell(cell.clone());
		}
		// Clones all the edges
		for (Iterator<Edge> iterator = edges.iterator(); iterator.hasNext(); )
		{
			Connection connection = (Connection) iterator.next();
			// It takes the vertex that corresponds to the start of the connection
			int startCellNumber = connection.getStartCell().getNumber();
			Cell startCell = space.getCell(startCellNumber);
			// It takes the vertex that corresponds to the end of the connection
			int endCellNumber = connection.getEndCell().getNumber();
			Cell endCell = space.getCell(endCellNumber);
			space.addConnection(new Connection(startCell, endCell, connection.getAction()));
		}

		return space;
	}

	// Treatment of object attributes
	/**
	 * It establish the bounds of the cells
	 * 
	 * If the lower bound is less than 2 then it's set to 2.
	 * If the upper bound is less to the lower bound then it's set to the lower bound.
	 */
	public void setCellsBounds(int minCells, int maxCells)
	{
		this.minCells = minCells < 2 ? 2 : minCells;
		this.maxCells = maxCells < this.minCells ? this.minCells : maxCells;
	}
	/**
	 * It establish the bounds of the actions
	 * 
	 * If the lower bound is less than 2 then it's set to 2.
	 * If the upper bound is less to the lower bound then it's set to the lower bound.
	 */
	public void setActionsBounds(int minActions, int maxActions)
	{
		this.minActions = minActions < 2 ? 2 : minActions;
		this.maxActions = maxActions < this.minActions ? this.minActions : maxActions;
	}
	public String getDescription()
	{
		return description;
	}
	/**
	 * It returns the cell that corresponds to the introduced number
	 * 
	 * If there not exists any cell with this number, it throws CellNotFoundException
	 */
	public Cell getCell(int number)
	{
		for (Vertex vertex : vertices.values())
		{
			Cell cell = (Cell) vertex;
			if (cell.getNumber() == number)
				return cell;
		}

		throw new CellNotFoundException(this);
	}
	public Collection<Cell> getCells()
	{
		Collection<Cell> cells = new ArrayList<Cell>();
		for (Vertex vertex : vertices.values())
			cells.add((Cell) vertex);
		return cells;
	}
	public int getNumberOfCells()
	{
		return getOrder();
	}
	public int getNumberOfActions()
	{
		return numberOfActions;
	}
	private void addCell(Cell cell)
	{
		addVertex(cell);
	}
	private void addConnection(Connection connection)
	{
		addEdge(connection);
	}

	// Object functions
	public boolean checkConsistency()
	{
		if (description == null)
			throw new RuntimeException("Space is not defined");
		return true;
	}
	/**
	 * Reset the space to its initial definition
	 */
	public void reset()
	{
		// It reset all the cells
		for (Iterator<Vertex> iterator = vertices.values().iterator(); iterator.hasNext(); )
		{
			Cell cell = (Cell) iterator.next();
			cell.reset();
		}
	}

	/**
	 * Generates a new topology of space
	 */
	public void generateSpace()
	{
		// It generates a new description of the space
		generateSpaceDescription();
		// It construct the space from the generated description
		createSpace();
	}
	/**
	 * Creates a new topology of space from the provided description
	 */
	public void constructSpace(String description)
	{
		this.description = description;
		createSpace();
	}
	/**
	 * Generates a new description of the space
	 */
	private void generateSpaceDescription()
	{
		IDistribution uniformDistribution = new UniformDistribution();
		IDistribution universalDistribution = new UniversalDistribution();
		// It decides the number of cells that will have the space
		int numCells = universalDistribution.getValue(minCells, maxCells);
		// It decides the number of actions that will have the space
		int maxNumActions = Math.min(numCells, maxActions);
		int numActions = universalDistribution.getValue(minActions, maxNumActions);

		description = "";
		// It generates all the cells
		for (int cell = 0; cell < numCells; cell++)
		{
			if (cell != 0)
				description += "|";
			// It generates all the actions for the cell
			for (int action = 1; action < numActions; action++)
			{
				// It decides to which cell will move with this action
				int movements = uniformDistribution.getValue(numCells);
				char movement = uniformDistribution.getValue(2) == 0 ? '+' : '-';
				description += action;
				for (int index = 0; index < movements; index++)
					description += movement;
			}
		}
	}
	/**
	 * Creates a new space following its description
	 */
	private void createSpace()
	{
		vertices.clear();
		edges.clear();

		// It takes the number of cells that has the described space
		StringTokenizer tokenizer = new StringTokenizer(description, "|");
		int numCells = tokenizer.countTokens();
		int numActions = 0;

		// It creates all the cells
		for (int cellNumber = 1; cellNumber <= numCells; cellNumber++)
		{
			Cell cell = new Cell(cellNumber);
			addCell(cell);
			addConnection(new Connection(cell, cell, 0));
		}

		// For each cell
		for (int cellNumber = 1; cellNumber <= numCells; cellNumber++)
		{
			String cellDescription = tokenizer.nextToken();
			int index = 0;
			while (index < cellDescription.length())
			{
				boolean exit = false;
				// Takes the action
				int action = (int) cellDescription.charAt(index) - 48;
				if (numActions < action)
					numActions = action;
				int destinationCell = cellNumber;
				// Takes the displacement
				while (!exit && ++index < cellDescription.length())
				{
					switch (cellDescription.charAt(index))
					{
					case '+':
						destinationCell++;
						break;
					case '-':
						destinationCell--;
						break;
					default:
						exit = true;
						break;
					}
				}
				destinationCell--;
				destinationCell = (destinationCell + numCells) % numCells;
				destinationCell++;

				// If the action goes to another cell add it as a new connection
				if (cellNumber != destinationCell)
					addConnection(new Connection(getCell(cellNumber), getCell(destinationCell), action));
			}
		}

		numberOfActions = numActions + 1;
	}
	/**
	 * Locates an Object in the numbered Cell
	 */
	public void locateObject(Objeto object, int cellNumber)
	{
		// Obtains the cell where the object will be placed
		Cell newCell = getCell(cellNumber);
		// Obtains the cell where the object was
		Cell oldCell = object.getLocation();
		// If the object was in a cell, remove it from there
		if (oldCell != null)
			oldCell.removeObject(object);
		// Assign to the object his new cell
		object.setLocation(newCell);
		// Add the object to the new cell
		newCell.addObject(object);
	}
}