package graph;


import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;


/**
 * It implements the interface of a graph in graph theory
 * 
 * @author Javier
 * @since 29/03/2011
 */
public class Graph implements IGraph
{
	// Object attributes
	/**
	 * Vertices that form a graph
	 */
	protected final Map<Integer, Vertex> vertices;
	/**
	 * Edges that form a graph
	 */
	protected final ArrayList<Edge> edges;


	// Constructors
	public Graph()
	{
		vertices = new HashMap<Integer, Vertex>();
		edges = new ArrayList<Edge>();
	}

	// Object overrides
	public Graph clone()
	{
		Graph graph = new Graph();

		// Clones all the vertices of the graph and add them to the new graph
		for (Vertex vertex : vertices.values())
			graph.addVertex(vertex.clone());

		// Clones all the edges of the graph and adds them to the new graph
		for (Edge edge : edges)
		{
			// It searches what vertex corresponds with the start of the edge
			Vertex startVertex = graph.getVertex(edge.getStartVertex().getIdentifier());
			// It searches what vertex corresponds with the end of the edge
			Vertex endVertex = graph.getVertex(edge.getEndVertex().getIdentifier());

			// It creates the new edge and adds it to the new graph
			Edge newEdge = new Edge(startVertex, endVertex);
			graph.addEdge(newEdge);
		}

		return graph;
	}

	// IGraph overrides
	public Collection<Vertex> getVertices()
	{
		return vertices.values();
	}
	public Vertex getVertex(int identifier)
	{
		return vertices.get(identifier);
	}
	public void addVertex(Vertex vertex)
	{
		if (vertex != null)
			vertices.put(vertex.getIdentifier(), vertex);
	}
	public void removeVertex(Vertex vertex)
	{
		vertices.remove(vertex.getIdentifier());
	}
	public Collection<Edge> getEdges()
	{
		return edges;
	}
	public void addEdge(Edge edge)
	{
		if (edge != null)
			edges.add(edge);
	}
	public void removeEdge(Edge edge)
	{
		edges.remove(edge);
	}
	/**
	 * It returns the number of vertices of the graph
	 */
	public int getOrder()
	{
		return vertices.size();
	}

	/**
	 * It checks if the graph is connected
	 */
	public boolean isConnected()
	{
		// Researched vertices during the check
		Collection<Vertex> reachedVertices = new ArrayList<Vertex>();
		// Auxiliar structure to visit the vertices
		Queue<Vertex> verticesQueue = new LinkedList<Vertex>();

		// It introduces the first vertex of the list from which it will start the checking
		verticesQueue.add(vertices.get(1));
		// It adds the fist vertex to the list of researched vertices
		reachedVertices.add(verticesQueue.peek());

		// While them exists vertices that have not been yet visited
		while (verticesQueue.size() > 0)
		{
			// Treat the next vertex
			Vertex actualVertex = verticesQueue.poll();
			// Treat all the adjacent vertices of the actual vertex
			for (Vertex adyacentVertex : actualVertex.getConnected())
			{
				// If the adjacent vertex is the own vertex then don't take it into account
				if (adyacentVertex.equals(actualVertex))
					continue;

				// If the adjacent vertex has already been reached don't take it into account
				if (reachedVertices.contains(adyacentVertex))
					continue;

				// Add the vertex to the list of vertices to visit
				verticesQueue.add(adyacentVertex);
				// Add the vertex to the list of reached vertices
				reachedVertices.add(adyacentVertex);
			}
		}

		// It returns if all vertex have been reached
		return reachedVertices.size() == vertices.size();
	}
	/**
	 * It checks if the graph is strongly connected
	 */
	public boolean isStrongConnected()
	{
		HashMap<Integer, Set<Integer>> cicles = new HashMap<Integer, Set<Integer>>();
		HashMap<Integer, Vertex> vertices = new HashMap<Integer, Vertex>();
		for (Vertex vertex : this.vertices.values())
			vertices.put(vertex.getIdentifier(), vertex);
		// Auxiliar structure to visit the vertices
		Queue<Vertex> verticesQueue = new LinkedList<Vertex>();

		// Introduces the first vertex of the list from which will start the checking
		verticesQueue.add(this.vertices.get(1));
		// It adds the first vertex to the list of reached vertex
		cicles.put(verticesQueue.peek().getIdentifier(), new LinkedHashSet<Integer>());

		// It prepares the data structures
		// While they exists vertex that have not yet been visited
		while (verticesQueue.size() > 0)
		{
			// Treat the next vertex
			Vertex actualVertex = verticesQueue.poll();
			// Treat all the adjacent vertices of the actual vertex
			for (Vertex adyacentVertex : actualVertex.getAdyacents())
			{
				// If the adjacent vertex is the own actual vertex don't take it into account
				if (adyacentVertex.equals(actualVertex))
					continue;

				Set<Integer> connections;
				// If the adjacent vertex has not yet been reached then prepare it
				if (!cicles.containsKey(adyacentVertex.getIdentifier()))
				{
					connections = new LinkedHashSet<Integer>();
					cicles.put(adyacentVertex.getIdentifier(), connections);
					verticesQueue.add(adyacentVertex);
				}
				else
					connections = cicles.get(adyacentVertex.getIdentifier());

				// Add a connection between the actual vertex and the adjacent
				connections.add(actualVertex.getIdentifier());
			}
		}

		// If all the vertices have not been reached, the graph is not strongly connected
		if (cicles.size() != this.vertices.size())
			return false;

		// Reduce the graph by reducing the number of cycles where is vertex 1
		Set<Integer> cicle = cicles.get(1);
		while (cicle.size() > 0)
		{
			// It takes a cycle
			Set<Integer> actual = cicle;
			Iterator<Integer> iterator = actual.iterator();
			// It takes the first available edge
			int vertexValue = vertices.get(iterator.next()).getIdentifier();;
			// It removes the taken edge
			iterator.remove();

			// Reduce to return to the current vertex
			while (vertexValue != 1)
			{
				// It is prepared to take the next vertex of the cycle
				actual = cicles.get(vertexValue);
				iterator = actual.iterator();

				// It merges the vertex to the cycle
				// It takes the first available edge
				int nextVertexValue = iterator.next();
				// It removes the taken edge
				iterator.remove();

				// It redirects the edges of the vertex to the cycle
				cicle.addAll(actual);
				// It removes the vertex
				cicles.remove(vertexValue);
				// Joints this vertex to the cycle 1
				vertices.put(vertexValue, vertices.get(1));

				// Pick the vertex pointed by the edge
				vertexValue = vertices.get(nextVertexValue).getIdentifier();
			}
		}

		// If it has rest only one cycle then the graph is strongly connected
		return cicles.size() == 1;
	}
}